# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: authority_keys.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14\x61uthority_keys.proto\x12!extensions.api.cast_channel.proto\"\x83\x01\n\rAuthorityKeys\x12\x42\n\x04keys\x18\x01 \x03(\x0b\x32\x34.extensions.api.cast_channel.proto.AuthorityKeys.Key\x1a.\n\x03Key\x12\x13\n\x0b\x66ingerprint\x18\x01 \x02(\x0c\x12\x12\n\npublic_key\x18\x02 \x02(\x0c\x42\x02H\x03')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'authority_keys_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'H\003'
  _globals['_AUTHORITYKEYS']._serialized_start=60
  _globals['_AUTHORITYKEYS']._serialized_end=191
  _globals['_AUTHORITYKEYS_KEY']._serialized_start=145
  _globals['_AUTHORITYKEYS_KEY']._serialized_end=191
# @@protoc_insertion_point(module_scope)



"""
Provides a controller for controlling the default media players
on the Chromecast.
"""

import abc
from dataclasses import dataclass
from functools import partial

from ..const import (
    CAST_TYPE_AUDIO,
    CAST_TYPE_CHROMECAST,
    CAST_TYPE_GROUP,
    MESSAGE_TYPE,
    REQUEST_ID,
    REQUEST_TIMEOUT,
    SESSION_ID,
)
from ..generated.cast_channel_pb2 import (  # pylint: disable=no-name-in-module
    CastMessage,
)
from ..response_handler import WaitResponse, chain_on_success
from . import BaseController, CallbackType

APP_ID = "appId"
ERROR_REASON = "reason"

NS_RECEIVER = "urn:x-cast:com.google.cast.receiver"

TYPE_GET_STATUS = "GET_STATUS"
TYPE_RECEIVER_STATUS = "RECEIVER_STATUS"
TYPE_LAUNCH = "LAUNCH"
TYPE_LAUNCH_ERROR = "LAUNCH_ERROR"

LAUNCH_CANCELLED = "CANCELLED"

VOLUME_CONTROL_TYPE_ATTENUATION = "attenuation"
VOLUME_CONTROL_TYPE_FIXED = "fixed"
VOLUME_CONTROL_TYPE_MASTER = "master"


@dataclass(frozen=True)
class CastStatus:
    """Cast status container."""

    is_active_input: bool | None
    is_stand_by: bool | None
    volume_level: float
    volume_muted: bool
    app_id: str | None
    display_name: str | None
    namespaces: list[str]
    session_id: str | None
    transport_id: str | None
    status_text: str
    icon_url: str | None
    volume_control_type: str


@dataclass(frozen=True)
class LaunchFailure:
    """Launch failure container."""

    reason: str | None
    app_id: str | None
    request_id: int | None


class CastStatusListener(abc.ABC):
    """Listener for receiving cast status events."""

    @abc.abstractmethod
    def new_cast_status(self, status: CastStatus) -> None:
        """Updated cast status."""


class LaunchErrorListener(abc.ABC):
    """Listener for receiving launch error events."""

    @abc.abstractmethod
    def new_launch_error(self, status: LaunchFailure) -> None:
        """Launch error."""


class ReceiverController(BaseController):
    """
    Controller to interact with the Chromecast platform.

    :param cast_type: Type of Chromecast device.
    """

    def __init__(self, cast_type: str = CAST_TYPE_CHROMECAST) -> None:
        super().__init__(NS_RECEIVER, target_platform=True)

        self.status: CastStatus | None = None
        self.launch_failure: LaunchFailure | None = None
        self.cast_type = cast_type

        self._status_listeners: list[CastStatusListener] = []
        self._launch_error_listeners: list[LaunchErrorListener] = []

    def disconnected(self) -> None:
        """Called when disconnected. Will erase status."""
        self.logger.info("Receiver:channel_disconnected")
        self.status = None

    @property
    def app_id(self) -> str | None:
        """Convenience method to retrieve current app id."""
        return self.status.app_id if self.status else None

    def receive_message(self, _message: CastMessage, data: dict) -> bool:
        """
        Called when a receiver message is received.

        data is message.payload_utf8 interpreted as a JSON dict.
        """
        if data[MESSAGE_TYPE] == TYPE_RECEIVER_STATUS:
            self._process_get_status(data)

            return True

        if data[MESSAGE_TYPE] == TYPE_LAUNCH_ERROR:
            self._process_launch_error(data)

            return True

        return False

    def register_status_listener(self, listener: CastStatusListener) -> None:
        """Register a status listener for when a new Chromecast status
        has been received. Listeners will be called with
        listener.new_cast_status(status)"""
        self._status_listeners.append(listener)

    def register_launch_error_listener(self, listener: LaunchErrorListener) -> None:
        """Register a listener for when a new launch error message
        has been received. Listeners will be called with
        listener.new_launch_error(launch_failure)"""
        self._launch_error_listeners.append(listener)

    def update_status(
        self,
        *,
        callback_function: CallbackType | None = None,
    ) -> None:
        """Sends a message to the Chromecast to update the status."""
        self.logger.debug("Receiver:Updating status")
        self.send_message(
            {MESSAGE_TYPE: TYPE_GET_STATUS}, callback_function=callback_function
        )

    def launch_app(
        self,
        app_id: str,
        *,
        force_launch: bool = False,
        callback_function: CallbackType | None = None,
    ) -> None:
        """Launches an app on the Chromecast.

        Will only launch if it is not currently running unless
        force_launch=True."""

        if not force_launch and self.status is None:
            self.update_status(
                callback_function=chain_on_success(
                    partial(self._send_launch_message, app_id, force_launch),
                    callback_function,
                )
            )

        else:
            self._send_launch_message(app_id, force_launch, callback_function)

    def _send_launch_message(
        self,
        app_id: str,
        force_launch: bool,
        callback_function: CallbackType | None,
        *,
        retry_on_cancelled_error: bool = True,
    ) -> None:
        if force_launch or self.app_id != app_id:
            self.logger.info("Receiver:Launching app %s", app_id)

            self.launch_failure = None

            def handle_launch_response(msg_sent: bool, response: dict | None) -> None:
                if (
                    msg_sent  # pylint: disable=too-many-boolean-expressions
                    and response
                    and response.get(MESSAGE_TYPE) == TYPE_LAUNCH_ERROR
                    and response.get(ERROR_REASON) == LAUNCH_CANCELLED
                    and not self._launch_error_listeners
                    and retry_on_cancelled_error
                ):
                    self.logger.info(
                        "Receiver:Launching app %s failed, retrying once", app_id
                    )
                    self._send_launch_message(
                        app_id,
                        force_launch,
                        callback_function,
                        retry_on_cancelled_error=False,
                    )
                    return

                if not callback_function:
                    return

                if not msg_sent or not response:
                    callback_function(False, response)
                    return

                if response[MESSAGE_TYPE] == TYPE_RECEIVER_STATUS:
                    callback_function(True, response)
                    return

                callback_function(False, response)

            self.send_message(
                {MESSAGE_TYPE: TYPE_LAUNCH, APP_ID: app_id},
                callback_function=handle_launch_response,
            )
        else:
            self.logger.info("Not launching app %s - already running", app_id)
            if callback_function:
                callback_function(True, None)

    def stop_app(
        self,
        *,
        callback_function: CallbackType | None = None,
    ) -> None:
        """Stops the current running app on the Chromecast."""
        self.logger.info("Receiver:Stopping current app '%s'", self.app_id)
        return self.send_message(
            {MESSAGE_TYPE: "STOP"},
            inc_session_id=True,
            callback_function=callback_function,
        )

    def set_volume(self, volume: float, timeout: float = REQUEST_TIMEOUT) -> float:
        """Allows to set volume. Should be value between 0..1.
        Returns the new volume.

        """
        volume = min(max(0, volume), 1)
        self.logger.info("Receiver:setting volume to %.2f", volume)
        response_handler = WaitResponse(timeout, "set volume")
        self.send_message(
            {MESSAGE_TYPE: "SET_VOLUME", "volume": {"level": volume}},
            callback_function=response_handler.callback,
        )
        response_handler.wait_response()
        return volume

    def set_volume_muted(self, muted: bool, timeout: float = REQUEST_TIMEOUT) -> None:
        """Allows to mute volume."""
        response_handler = WaitResponse(timeout, "mute volume")
        self.send_message(
            {MESSAGE_TYPE: "SET_VOLUME", "volume": {"muted": muted}},
            callback_function=response_handler.callback,
        )
        response_handler.wait_response()

    @staticmethod
    def _parse_status(data: dict, cast_type: str) -> CastStatus:
        """
        Parses a STATUS message and returns a CastStatus object.

        :type data: dict
        :param cast_type: Type of Chromecast.
        :rtype: CastStatus
        """
        status_data: dict = data.get("status", {})

        volume_data: dict = status_data.get("volume", {})

        try:
            app_data: dict = status_data["applications"][0]
        except (KeyError, IndexError):
            app_data = {}

        is_audio = cast_type in (CAST_TYPE_AUDIO, CAST_TYPE_GROUP)

        status = CastStatus(
            data.get("isActiveInput", None if is_audio else False),
            data.get("isStandBy", None if is_audio else True),
            volume_data.get("level", 1.0),
            volume_data.get("muted", False),
            app_data.get(APP_ID),
            app_data.get("displayName"),
            [item["name"] for item in app_data.get("namespaces", [])],
            app_data.get(SESSION_ID),
            app_data.get("transportId"),
            app_data.get("statusText", ""),
            app_data.get("iconUrl"),
            volume_data.get("controlType", VOLUME_CONTROL_TYPE_ATTENUATION),
        )
        return status

    def _process_get_status(self, data: dict) -> None:
        """Processes a received STATUS message and notifies listeners."""
        status = self._parse_status(data, self.cast_type)
        self.status = status

        self.logger.debug("Received status: %s", self.status)

        for listener in self._status_listeners:
            try:
                listener.new_cast_status(self.status)
            except Exception:  # pylint: disable=broad-except
                self.logger.exception(
                    "Exception thrown when calling cast status listener"
                )

    @staticmethod
    def _parse_launch_error(data: dict) -> LaunchFailure:
        """
        Parses a LAUNCH_ERROR message and returns a LaunchFailure object.

        :type data: dict
        :rtype: LaunchFailure
        """
        return LaunchFailure(
            data.get(ERROR_REASON, None), data.get(APP_ID), data.get(REQUEST_ID)
        )

    def _process_launch_error(self, data: dict) -> None:
        """
        Processes a received LAUNCH_ERROR message and notifies listeners.
        """
        launch_failure = self._parse_launch_error(data)
        self.launch_failure = launch_failure

        self.logger.debug("Launch status: %s", launch_failure)

        for listener in self._launch_error_listeners:
            try:
                listener.new_launch_error(launch_failure)
            except Exception:  # pylint: disable=broad-except
                self.logger.exception(
                    "Exception thrown when calling launch error listener"
                )

    def tear_down(self) -> None:
        """Called when controller is destroyed."""
        super().tear_down()

        self.status = None
        self.launch_failure = None

        self._status_listeners = []



"""
Data and methods to retrieve app specific configuration
"""

import json
from typing import cast

import requests

APP_BACKDROP = "E8C28D3C"
APP_YOUTUBE = "233637DE"
APP_MEDIA_RECEIVER = "CC1AD845"
APP_PLEX = "06ee44ee-e7e3-4249-83b6-f5d0b6f07f34_1"
APP_DASHCAST = "84912283"
APP_HOMEASSISTANT_LOVELACE = "A078F6B0"
APP_HOMEASSISTANT_MEDIA = "B45F4572"
APP_SUPLA = "A41B766D"
APP_YLEAREENA = "A9BCCB7C"
APP_BUBBLEUPNP = "3927FA74"
APP_BBCSOUNDS = "D350F6A9"
APP_BBCIPLAYER = "5E81F6DB"
APP_SHAKA = "07AEE832"
APP_NRKTV = "3AEDF8D1"
APP_NRKRADIO = "A49874B1"
APP_AUDIBLE = "25456794"


def get_possible_app_ids() -> list[str]:
    """Returns all possible app ids."""

    try:
        req = requests.get(
            "https://clients3.google.com/cast/chromecast/device/baseconfig",
            timeout=10,
        )
        data = json.loads(req.text[4:])

        return cast(
            list[str],
            [app["app_id"] for app in data["applications"]] + data["enabled_app_ids"],
        )

    except ValueError:
        # If json fails to parse
        return []


def get_app_config(app_id: str) -> dict:
    """Get specific configuration for 'app_id'."""
    try:
        req = requests.get(
            f"https://clients3.google.com/cast/chromecast/device/app?a={app_id}",
            timeout=10,
        )

        return cast(dict, json.loads(req.text[4:])) if req.status_code == 200 else {}

    except ValueError:
        # If json fails to parse
        return {}



""" Choose a controller and quick play """

from typing import Any

from . import Chromecast
from .controllers import QuickPlayController
from .controllers.bbciplayer import BbcIplayerController
from .controllers.bbcsounds import BbcSoundsController
from .controllers.bubbleupnp import BubbleUPNPController
from .controllers.homeassistant_media import HomeAssistantMediaController
from .controllers.media import DefaultMediaReceiverController
from .controllers.supla import SuplaController
from .controllers.yleareena import YleAreenaController
from .controllers.youtube import YouTubeController
from .controllers.shaka import ShakaController
from .controllers.nrktv import NrkTvController
from .controllers.nrkradio import NrkRadioController


def quick_play(  # pylint:disable=too-many-branches
    cast: Chromecast,
    app_name: str,
    data: dict[str, Any],
    timeout: float = 30.0,
) -> None:
    """
    Given a Chromecast connection, launch the app `app_name` and start playing media
    based on parameters defined in `data`.

    :param cast: Chromecast connection to cast to
    :param app_name: App name "slug" to cast
    :param data: Data to send to the app controller. Must contain "media_id", and other
        values can be passed depending on the controller.
    :type cast: Chromecast
    :type app_name: string
    :type data: dict

    `data` can contain the following keys:
        media_id: string (Required)
            Primary identifier of the media
        media_type: string
            Type of the media identified by `media_id`. e.g. "program" if the media is a
            program name instead of a direct item id.
            When using a regular media controller (e.g. BubbleUPNP) this should be the
            content_type ('audio/mp3')
        enqueue: boolean
            Enqueue the media to the current playlist, if possible.
        index: string
            Play index x of matching media. "random" should also be allowed.
        audio_lang: string
            Audio language (3 characters for YleAreena)
        text_lang: string
            Subtitle language (3 characters for YleAreena)

    Youtube-specific:
        playlist_id: string
            Youtube playlist id

    Supla-specific:
        is_live: boolean
            Whether the media is a livestream

    Media controller (BubbleUPNP)-specific:
        stream_type: string
            "BUFFERED" or "LIVE"
    """

    controller: QuickPlayController
    if app_name == "bbciplayer":
        controller = BbcIplayerController()
    elif app_name == "bbcsounds":
        controller = BbcSoundsController()
    elif app_name == "bubbleupnp":
        controller = BubbleUPNPController()
    elif app_name == "default_media_receiver":
        controller = DefaultMediaReceiverController()
    elif app_name == "homeassistant_media":
        controller = HomeAssistantMediaController()
    elif app_name == "supla":
        controller = SuplaController()
    elif app_name == "yleareena":
        controller = YleAreenaController()
    elif app_name == "youtube":
        controller = YouTubeController()
    elif app_name == "shaka":
        controller = ShakaController()
    elif app_name == "nrktv":
        controller = NrkTvController()
    elif app_name == "nrkradio":
        controller = NrkRadioController()
    else:
        raise NotImplementedError()

    cast.register_handler(controller)

    try:
        controller.quick_play(**data, timeout=timeout)
    finally:
        cast.unregister_handler(controller)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cast_channel.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12\x63\x61st_channel.proto\x12\x1b\x65xtensions.api.cast_channel\"\xe3\x02\n\x0b\x43\x61stMessage\x12R\n\x10protocol_version\x18\x01 \x02(\x0e\x32\x38.extensions.api.cast_channel.CastMessage.ProtocolVersion\x12\x11\n\tsource_id\x18\x02 \x02(\t\x12\x16\n\x0e\x64\x65stination_id\x18\x03 \x02(\t\x12\x11\n\tnamespace\x18\x04 \x02(\t\x12J\n\x0cpayload_type\x18\x05 \x02(\x0e\x32\x34.extensions.api.cast_channel.CastMessage.PayloadType\x12\x14\n\x0cpayload_utf8\x18\x06 \x01(\t\x12\x16\n\x0epayload_binary\x18\x07 \x01(\x0c\"!\n\x0fProtocolVersion\x12\x0e\n\nCASTV2_1_0\x10\x00\"%\n\x0bPayloadType\x12\n\n\x06STRING\x10\x00\x12\n\n\x06\x42INARY\x10\x01\"\xce\x01\n\rAuthChallenge\x12]\n\x13signature_algorithm\x18\x01 \x01(\x0e\x32/.extensions.api.cast_channel.SignatureAlgorithm:\x0fRSASSA_PKCS1v15\x12\x14\n\x0csender_nonce\x18\x02 \x01(\x0c\x12H\n\x0ehash_algorithm\x18\x03 \x01(\x0e\x32*.extensions.api.cast_channel.HashAlgorithm:\x04SHA1\"\xb0\x02\n\x0c\x41uthResponse\x12\x11\n\tsignature\x18\x01 \x02(\x0c\x12\x1f\n\x17\x63lient_auth_certificate\x18\x02 \x02(\x0c\x12 \n\x18intermediate_certificate\x18\x03 \x03(\x0c\x12]\n\x13signature_algorithm\x18\x04 \x01(\x0e\x32/.extensions.api.cast_channel.SignatureAlgorithm:\x0fRSASSA_PKCS1v15\x12\x14\n\x0csender_nonce\x18\x05 \x01(\x0c\x12H\n\x0ehash_algorithm\x18\x06 \x01(\x0e\x32*.extensions.api.cast_channel.HashAlgorithm:\x04SHA1\x12\x0b\n\x03\x63rl\x18\x07 \x01(\x0c\"\xa3\x01\n\tAuthError\x12\x44\n\nerror_type\x18\x01 \x02(\x0e\x32\x30.extensions.api.cast_channel.AuthError.ErrorType\"P\n\tErrorType\x12\x12\n\x0eINTERNAL_ERROR\x10\x00\x12\n\n\x06NO_TLS\x10\x01\x12#\n\x1fSIGNATURE_ALGORITHM_UNAVAILABLE\x10\x02\"\xc6\x01\n\x11\x44\x65viceAuthMessage\x12=\n\tchallenge\x18\x01 \x01(\x0b\x32*.extensions.api.cast_channel.AuthChallenge\x12;\n\x08response\x18\x02 \x01(\x0b\x32).extensions.api.cast_channel.AuthResponse\x12\x35\n\x05\x65rror\x18\x03 \x01(\x0b\x32&.extensions.api.cast_channel.AuthError*J\n\x12SignatureAlgorithm\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x13\n\x0fRSASSA_PKCS1v15\x10\x01\x12\x0e\n\nRSASSA_PSS\x10\x02*%\n\rHashAlgorithm\x12\x08\n\x04SHA1\x10\x00\x12\n\n\x06SHA256\x10\x01\x42\x02H\x03')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'cast_channel_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'H\003'
  _globals['_SIGNATUREALGORITHM']._serialized_start=1292
  _globals['_SIGNATUREALGORITHM']._serialized_end=1366
  _globals['_HASHALGORITHM']._serialized_start=1368
  _globals['_HASHALGORITHM']._serialized_end=1405
  _globals['_CASTMESSAGE']._serialized_start=52
  _globals['_CASTMESSAGE']._serialized_end=407
  _globals['_CASTMESSAGE_PROTOCOLVERSION']._serialized_start=335
  _globals['_CASTMESSAGE_PROTOCOLVERSION']._serialized_end=368
  _globals['_CASTMESSAGE_PAYLOADTYPE']._serialized_start=370
  _globals['_CASTMESSAGE_PAYLOADTYPE']._serialized_end=407
  _globals['_AUTHCHALLENGE']._serialized_start=410
  _globals['_AUTHCHALLENGE']._serialized_end=616
  _globals['_AUTHRESPONSE']._serialized_start=619
  _globals['_AUTHRESPONSE']._serialized_end=923
  _globals['_AUTHERROR']._serialized_start=926
  _globals['_AUTHERROR']._serialized_end=1089
  _globals['_AUTHERROR_ERRORTYPE']._serialized_start=1009
  _globals['_AUTHERROR_ERRORTYPE']._serialized_end=1089
  _globals['_DEVICEAUTHMESSAGE']._serialized_start=1092
  _globals['_DEVICEAUTHMESSAGE']._serialized_end=1290
# @@protoc_insertion_point(module_scope)



"""
Chromecast types
"""

from __future__ import annotations

import asyncio
from dataclasses import dataclass
from uuid import UUID

import zeroconf

ZEROCONF_ERRORS: tuple[type[Exception], ...] = (IOError, asyncio.TimeoutError)
if hasattr(zeroconf, "EventLoopBlocked"):
    # Added in zeroconf 0.37.0
    ZEROCONF_ERRORS = (*ZEROCONF_ERRORS, zeroconf.EventLoopBlocked)


@dataclass(frozen=True)
class CastInfo:
    """Cast info container."""

    services: set[HostServiceInfo | MDNSServiceInfo]
    uuid: UUID
    model_name: str | None
    friendly_name: str | None
    host: str
    port: int
    cast_type: str | None
    manufacturer: str | None


@dataclass(frozen=True)
class HostServiceInfo:
    """Service info container."""

    host: str
    port: int


@dataclass(frozen=True)
class MDNSServiceInfo:
    """Service info container."""

    name: str



"""Controller to send and respond to heartbeat messages."""

from __future__ import annotations

import logging
import time

from . import BaseController

from ..const import MESSAGE_TYPE, PLATFORM_DESTINATION_ID
from ..error import ControllerNotRegistered, NotConnected, PyChromecastStopped

# pylint: disable-next=no-name-in-module
from ..generated.cast_channel_pb2 import CastMessage

NS_HEARTBEAT = "urn:x-cast:com.google.cast.tp.heartbeat"

TYPE_PING = "PING"
TYPE_PONG = "PONG"

HB_PING_TIME = 10
HB_PONG_TIME = 10


class HeartbeatController(BaseController):
    """Controller to send and respond to heartbeat messages."""

    def __init__(self) -> None:
        super().__init__(NS_HEARTBEAT, target_platform=True)
        self.last_ping = 0.0
        self.last_pong = time.time()
        self.logger = logging.getLogger(__name__)

    def receive_message(self, _message: CastMessage, data: dict) -> bool:
        """
        Called when a heartbeat message is received.

        data is message.payload_utf8 interpreted as a JSON dict.
        """
        if self._socket_client is None:
            raise ControllerNotRegistered

        if self._socket_client.is_stopped:
            return True

        if data[MESSAGE_TYPE] == TYPE_PING:
            try:
                self._socket_client.send_message(
                    PLATFORM_DESTINATION_ID,
                    self.namespace,
                    {MESSAGE_TYPE: TYPE_PONG},
                    no_add_request_id=True,
                )
            except PyChromecastStopped:
                self._socket_client.logger.debug(
                    "Heartbeat error when sending response, "
                    "Chromecast connection has stopped"
                )

            return True

        if data[MESSAGE_TYPE] == TYPE_PONG:
            self.reset()
            return True

        return False

    def ping(self) -> None:
        """Send a ping message."""
        if self._socket_client is None:
            raise ControllerNotRegistered

        self.last_ping = time.time()
        try:
            self.send_message({MESSAGE_TYPE: TYPE_PING})
        except NotConnected:
            self._socket_client.logger.error(
                "Chromecast is disconnected. Cannot ping until reconnected."
            )

    def reset(self) -> None:
        """Reset expired counter."""
        self.last_pong = time.time()

    def is_expired(self) -> bool:
        """Indicates if connection has expired."""
        if time.time() - self.last_ping > HB_PING_TIME:
            self.ping()

        return (time.time() - self.last_pong) > HB_PING_TIME + HB_PONG_TIME



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class SignatureAlgorithm(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    UNSPECIFIED: _ClassVar[SignatureAlgorithm]
    RSASSA_PKCS1v15: _ClassVar[SignatureAlgorithm]
    RSASSA_PSS: _ClassVar[SignatureAlgorithm]

class HashAlgorithm(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    SHA1: _ClassVar[HashAlgorithm]
    SHA256: _ClassVar[HashAlgorithm]
UNSPECIFIED: SignatureAlgorithm
RSASSA_PKCS1v15: SignatureAlgorithm
RSASSA_PSS: SignatureAlgorithm
SHA1: HashAlgorithm
SHA256: HashAlgorithm

class CastMessage(_message.Message):
    __slots__ = ("protocol_version", "source_id", "destination_id", "namespace", "payload_type", "payload_utf8", "payload_binary")
    class ProtocolVersion(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        CASTV2_1_0: _ClassVar[CastMessage.ProtocolVersion]
    CASTV2_1_0: CastMessage.ProtocolVersion
    class PayloadType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        STRING: _ClassVar[CastMessage.PayloadType]
        BINARY: _ClassVar[CastMessage.PayloadType]
    STRING: CastMessage.PayloadType
    BINARY: CastMessage.PayloadType
    PROTOCOL_VERSION_FIELD_NUMBER: _ClassVar[int]
    SOURCE_ID_FIELD_NUMBER: _ClassVar[int]
    DESTINATION_ID_FIELD_NUMBER: _ClassVar[int]
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_TYPE_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_UTF8_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_BINARY_FIELD_NUMBER: _ClassVar[int]
    protocol_version: CastMessage.ProtocolVersion
    source_id: str
    destination_id: str
    namespace: str
    payload_type: CastMessage.PayloadType
    payload_utf8: str
    payload_binary: bytes
    def __init__(self, protocol_version: _Optional[_Union[CastMessage.ProtocolVersion, str]] = ..., source_id: _Optional[str] = ..., destination_id: _Optional[str] = ..., namespace: _Optional[str] = ..., payload_type: _Optional[_Union[CastMessage.PayloadType, str]] = ..., payload_utf8: _Optional[str] = ..., payload_binary: _Optional[bytes] = ...) -> None: ...

class AuthChallenge(_message.Message):
    __slots__ = ("signature_algorithm", "sender_nonce", "hash_algorithm")
    SIGNATURE_ALGORITHM_FIELD_NUMBER: _ClassVar[int]
    SENDER_NONCE_FIELD_NUMBER: _ClassVar[int]
    HASH_ALGORITHM_FIELD_NUMBER: _ClassVar[int]
    signature_algorithm: SignatureAlgorithm
    sender_nonce: bytes
    hash_algorithm: HashAlgorithm
    def __init__(self, signature_algorithm: _Optional[_Union[SignatureAlgorithm, str]] = ..., sender_nonce: _Optional[bytes] = ..., hash_algorithm: _Optional[_Union[HashAlgorithm, str]] = ...) -> None: ...

class AuthResponse(_message.Message):
    __slots__ = ("signature", "client_auth_certificate", "intermediate_certificate", "signature_algorithm", "sender_nonce", "hash_algorithm", "crl")
    SIGNATURE_FIELD_NUMBER: _ClassVar[int]
    CLIENT_AUTH_CERTIFICATE_FIELD_NUMBER: _ClassVar[int]
    INTERMEDIATE_CERTIFICATE_FIELD_NUMBER: _ClassVar[int]
    SIGNATURE_ALGORITHM_FIELD_NUMBER: _ClassVar[int]
    SENDER_NONCE_FIELD_NUMBER: _ClassVar[int]
    HASH_ALGORITHM_FIELD_NUMBER: _ClassVar[int]
    CRL_FIELD_NUMBER: _ClassVar[int]
    signature: bytes
    client_auth_certificate: bytes
    intermediate_certificate: _containers.RepeatedScalarFieldContainer[bytes]
    signature_algorithm: SignatureAlgorithm
    sender_nonce: bytes
    hash_algorithm: HashAlgorithm
    crl: bytes
    def __init__(self, signature: _Optional[bytes] = ..., client_auth_certificate: _Optional[bytes] = ..., intermediate_certificate: _Optional[_Iterable[bytes]] = ..., signature_algorithm: _Optional[_Union[SignatureAlgorithm, str]] = ..., sender_nonce: _Optional[bytes] = ..., hash_algorithm: _Optional[_Union[HashAlgorithm, str]] = ..., crl: _Optional[bytes] = ...) -> None: ...

class AuthError(_message.Message):
    __slots__ = ("error_type",)
    class ErrorType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        INTERNAL_ERROR: _ClassVar[AuthError.ErrorType]
        NO_TLS: _ClassVar[AuthError.ErrorType]
        SIGNATURE_ALGORITHM_UNAVAILABLE: _ClassVar[AuthError.ErrorType]
    INTERNAL_ERROR: AuthError.ErrorType
    NO_TLS: AuthError.ErrorType
    SIGNATURE_ALGORITHM_UNAVAILABLE: AuthError.ErrorType
    ERROR_TYPE_FIELD_NUMBER: _ClassVar[int]
    error_type: AuthError.ErrorType
    def __init__(self, error_type: _Optional[_Union[AuthError.ErrorType, str]] = ...) -> None: ...

class DeviceAuthMessage(_message.Message):
    __slots__ = ("challenge", "response", "error")
    CHALLENGE_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    challenge: AuthChallenge
    response: AuthResponse
    error: AuthError
    def __init__(self, challenge: _Optional[_Union[AuthChallenge, _Mapping]] = ..., response: _Optional[_Union[AuthResponse, _Mapping]] = ..., error: _Optional[_Union[AuthError, _Mapping]] = ...) -> None: ...



"""
Controller to interface with Supla.
"""

import logging
from typing import Any

from . import CallbackType, QuickPlayController
from ..config import APP_SUPLA
from ..response_handler import WaitResponse

APP_NAMESPACE = "urn:x-cast:fi.ruutu.chromecast"


# pylint: disable=too-many-instance-attributes
class SuplaController(QuickPlayController):
    """Controller to interact with Supla namespace."""

    def __init__(self) -> None:
        super().__init__(APP_NAMESPACE, APP_SUPLA)

        self.logger = logging.getLogger(__name__)

    def play_media(
        self,
        media_id: str,
        *,
        is_live: bool = False,
        callback_function: CallbackType | None = None,
    ) -> None:
        """
        Play Supla media
        """
        msg = {
            "type": "load",
            "mediaId": media_id,
            "currentTime": 0,
            "isLive": is_live,
            "isAtLiveMoment": False,
            "bookToken": "",
            "sample": True,
            "fw_site": "Supla",
            "Sanoma_adkv": "",
            "prerollAdsPlayed": True,
            "supla": True,
            "nextInSequenceList": 0,
            "playbackRate": 1,
            "env": "prod",
        }

        self.send_message(
            msg,
            inc_session_id=False,
            callback_function=callback_function,
            no_add_request_id=True,
        )

    def quick_play(
        self, *, media_id: str, timeout: float, is_live: bool = False, **kwargs: Any
    ) -> None:
        """Quick Play"""
        response_handler = WaitResponse(timeout, f"supla quick play {media_id}")
        self.play_media(
            media_id,
            is_live=is_live,
            **kwargs,
            callback_function=response_handler.callback,
        )
        response_handler.wait_response()



"""
Provides controllers to handle specific namespaces in Chromecast communication.
"""

from __future__ import annotations

import abc
from functools import partial
import logging
from typing import TYPE_CHECKING, Any, Protocol

from ..error import UnsupportedNamespace, ControllerNotRegistered
from ..generated.cast_channel_pb2 import (  # pylint: disable=no-name-in-module
    CastMessage,
)
from ..response_handler import CallbackType, chain_on_success

if TYPE_CHECKING:
    from ..socket_client import SocketClient


class SendMessageFunc(Protocol):
    """Protocol for SocketClient's send message functions."""

    def __call__(
        self,
        namespace: str,
        message: Any,
        *,
        inc_session_id: bool = False,
        callback_function: CallbackType | None = None,
        no_add_request_id: bool = False,
    ) -> None: ...


class BaseController(abc.ABC):
    """ABC for namespace controllers."""

    def __init__(
        self,
        namespace: str,
        supporting_app_id: str | None = None,
        target_platform: bool = False,
        app_must_match: bool = False,
    ) -> None:
        """
        Initialize the controller.

        namespace:         the namespace this controller will act on
        supporting_app_id: app to be launched if app is running with
                           unsupported namespace.
        target_platform:   set to True if you target the platform instead of
                           current app.
        app_must_match:    set to True if the app should be launched even if the
                           namespace is supported by another app.
        """
        self.app_must_match = app_must_match
        self.namespace = namespace
        self.supporting_app_id = supporting_app_id
        self.target_platform = target_platform

        self._socket_client: SocketClient | None = None
        self._message_func: SendMessageFunc | None = None

        self.logger = logging.getLogger(__name__)

    @property
    def is_active(self) -> bool:
        """True if the controller is connected to a socket client and the
        Chromecast is running an app that supports this controller."""
        return (
            self._socket_client is not None
            and self.namespace in self._socket_client.app_namespaces
        )

    def launch(
        self,
        *,
        callback_function: CallbackType | None = None,
        force_launch: bool = False,
    ) -> None:
        """If set, launches app related to the controller."""
        if self.supporting_app_id is None:
            self.logger.debug(
                "%s: Can't launch app with no supporting app_id",
                self.__class__.__name__,
            )
            if callback_function:
                callback_function(False, None)
            return

        if self._socket_client is None:
            if callback_function:
                callback_function(False, None)
            raise ControllerNotRegistered

        self._socket_client.receiver_controller.launch_app(
            self.supporting_app_id,
            force_launch=force_launch,
            callback_function=callback_function,
        )

    def registered(self, socket_client: SocketClient) -> None:
        """Called when a controller is registered."""
        self._socket_client = socket_client

        if self.target_platform:
            self._message_func = self._socket_client.send_platform_message
        else:
            self._message_func = self._socket_client.send_app_message

    def unregistered(self) -> None:
        """Called when a controller is unregistered."""
        self._message_func = None

    def channel_connected(self) -> None:
        """Called when a channel has been openend that supports the
        namespace of this controller."""

    def channel_disconnected(self) -> None:
        """Called when a channel is disconnected."""

    def send_message(
        self,
        data: Any,
        *,
        inc_session_id: bool = False,
        callback_function: CallbackType | None = None,
        no_add_request_id: bool = False,
    ) -> None:
        """
        Send a message on this namespace to the Chromecast. Ensures app is loaded.

        Will raise a NotConnected exception if not connected.
        """
        if self._socket_client is None:
            if callback_function:
                callback_function(False, None)
            raise ControllerNotRegistered

        receiver_ctrl = self._socket_client.receiver_controller

        if not self.target_platform and (
            self.namespace not in self._socket_client.app_namespaces
            or (self.app_must_match and receiver_ctrl.app_id != self.supporting_app_id)
        ):
            if self.supporting_app_id is not None:
                self.launch(
                    callback_function=chain_on_success(
                        partial(
                            self.send_message_nocheck,
                            data,
                            inc_session_id=inc_session_id,
                            no_add_request_id=no_add_request_id,
                        ),
                        callback_function,
                    )
                )
                return

            if callback_function:
                callback_function(False, None)
            raise UnsupportedNamespace(
                f"Namespace {self.namespace} is not supported by running application."
            )

        self.send_message_nocheck(
            data,
            inc_session_id=inc_session_id,
            callback_function=callback_function,
            no_add_request_id=no_add_request_id,
        )

    def send_message_nocheck(
        self,
        data: Any,
        *,
        inc_session_id: bool = False,
        callback_function: CallbackType | None = None,
        no_add_request_id: bool = False,
    ) -> None:
        """Send a message."""
        if TYPE_CHECKING:
            assert self._message_func

        self._message_func(
            self.namespace,
            data,
            inc_session_id=inc_session_id,
            callback_function=callback_function,
            no_add_request_id=no_add_request_id,
        )

    def receive_message(self, _message: CastMessage, _data: dict) -> bool:
        """
        Called when a message is received that matches the namespace.
        Returns boolean indicating if message was handled.
        data is message.payload_utf8 interpreted as a JSON dict.
        """
        return False

    def tear_down(self) -> None:
        """Called when we are shutting down."""
        self._socket_client = None
        self._message_func = None


class QuickPlayController(BaseController, abc.ABC):
    """ABC for controller which supports quick play."""

    @abc.abstractmethod
    def quick_play(self, *, media_id: str, timeout: float, **kwargs: Any) -> None:
        """Quick Play support for a controller."""



"""
Controller to interface with Home Assistant
"""

from collections.abc import Callable
from functools import partial
import logging
import threading
from typing import Any

from ..config import APP_HOMEASSISTANT_LOVELACE
from ..error import PyChromecastError
from ..generated.cast_channel_pb2 import (  # pylint: disable=no-name-in-module
    CastMessage,
)
from ..response_handler import chain_on_success
from . import CallbackType, BaseController


APP_NAMESPACE = "urn:x-cast:com.nabucasa.hast"
DEFAULT_HASS_CONNECT_TIMEOUT = 30

# Error codes sent in receiver_error messages
ERR_CONNECTION_FAILED = 1
ERR_AUTHENTICATION_FAILED = 2
ERR_CONNECTION_LOST = 3
ERR_HASS_URL_MISSING = 4
ERR_NO_HTTPS = 5
ERR_WRONG_INSTANCE = 20
ERR_NOT_CONNECTED = 21
ERR_FETCH_CONFIG_FAILED = 22

_LOGGER = logging.getLogger(__name__)


class HomeAssistantController(BaseController):
    """Controller to interact with Home Assistant."""

    def __init__(
        self,
        *,
        hass_url: str,
        hass_uuid: str,
        client_id: str | None,
        refresh_token: str,
        unregister: Callable[[], None],
        app_namespace: str = APP_NAMESPACE,
        app_id: str = APP_HOMEASSISTANT_LOVELACE,
        hass_connect_timeout: float = DEFAULT_HASS_CONNECT_TIMEOUT,
    ) -> None:
        _LOGGER.debug("HomeAssistantController.__init__")
        super().__init__(app_namespace, app_id)
        self.hass_url = hass_url
        self.hass_uuid = hass_uuid
        self.client_id = client_id
        self.refresh_token = refresh_token
        self.unregister = unregister
        self.hass_connect_timeout = hass_connect_timeout
        # {
        #   connected: boolean;
        #   showDemo: boolean;
        #   hassUrl?: string;
        #   lovelacePath?: string | number | null;
        #   urlPath?: string | null;
        # }
        self.status: dict | None = None
        self._hass_connecting_event = threading.Event()
        self._hass_connecting_event.set()
        self._on_connect: list[CallbackType] = []

    @property
    def hass_connected(self) -> bool:
        """Return if connected to Home Assistant."""
        return (
            self.status is not None
            and self.status["connected"]
            and self.status["hassUrl"] == self.hass_url
            and self.status["hassUUID"] == self.hass_uuid
        )

    def channel_connected(self) -> None:
        """Called when a channel has been openend that supports the
        namespace of this controller."""
        _LOGGER.debug("HomeAssistantController.channel_connected")
        self.get_status()

    def channel_disconnected(self) -> None:
        """Called when a channel is disconnected."""
        _LOGGER.debug("HomeAssistantController.channel_disconnected")
        self.status = None
        self._hass_connecting_event.set()

    def receive_message(self, _message: CastMessage, data: dict) -> bool:
        """Called when a message is received."""
        if data.get("type") == "receiver_status":
            _LOGGER.debug("HomeAssistantController.receive_message %s", data)
            if data["hassUrl"] != self.hass_url or data["hassUUID"] != self.hass_uuid:
                self.logger.info("Received status for another instance")
                self.unregister()
                return True

            was_connected = self.hass_connected
            self.status = data

            if was_connected or not self.hass_connected:
                _LOGGER.debug(
                    "HomeAssistantController.receive_message already connected"
                )
                return True

            # We just got connected, call the callbacks.
            self._hass_connecting_event.set()
            self._call_on_connect_callbacks(True)
            return True

        if data.get("type") == "receiver_error":
            _LOGGER.debug("HomeAssistantController.receive_message %s", data)
            if data.get("error_code") == ERR_WRONG_INSTANCE:
                self.logger.info("Received ERR_WRONG_INSTANCE")
                self.unregister()
            return True

        return False

    def _call_on_connect_callbacks(self, msg_sent: bool) -> None:
        """Call on connect callbacks."""
        _LOGGER.debug("HomeAssistantController._call_on_connect_callbacks %s", msg_sent)
        while self._on_connect:
            self._on_connect.pop()(msg_sent, None)

    def _connect_hass(self, callback_function: CallbackType) -> None:
        """Connect to Home Assistant and call the provided callback."""
        _LOGGER.debug("HomeAssistantController._connect_hass")
        self._on_connect.append(callback_function)

        if not self._hass_connecting_event.is_set():
            _LOGGER.debug(
                "HomeAssistantController._connect_hass _hass_connecting_event not set"
            )
            return

        self._hass_connecting_event.clear()
        try:
            self.send_message(
                {
                    "type": "connect",
                    "refreshToken": self.refresh_token,
                    "clientId": self.client_id,
                    "hassUrl": self.hass_url,
                    "hassUUID": self.hass_uuid,
                }
            )
        except Exception:  # pylint: disable=broad-except
            _LOGGER.debug(
                "HomeAssistantController._connect_hass failed to send connect message"
            )
            self._hass_connecting_event.set()
            self._call_on_connect_callbacks(False)
            raise

        self._hass_connecting_event.wait(self.hass_connect_timeout)
        try:
            if not self._hass_connecting_event.is_set():
                self.logger.warning("_connect_hass failed for %s", self.hass_url)
                self._call_on_connect_callbacks(False)
                raise PyChromecastError()  # pylint: disable=broad-exception-raised
        finally:
            self._hass_connecting_event.set()

    def show_demo(self) -> None:
        """Show the demo."""
        self.send_message({"type": "show_demo"})

    def get_status(self, *, callback_function: CallbackType | None = None) -> None:
        """Get status of Home Assistant Cast."""
        _LOGGER.debug("HomeAssistantController.get_status")
        self._send_connected_message(
            {
                "type": "get_status",
                "hassUrl": self.hass_url,
                "hassUUID": self.hass_uuid,
            },
            callback_function=callback_function,
        )

    def show_lovelace_view(
        self,
        view_path: str | int | None,
        url_path: str | None = None,
        *,
        callback_function: CallbackType | None = None,
    ) -> None:
        """Show a Lovelace UI."""
        _LOGGER.debug("HomeAssistantController.show_lovelace_view")
        self._send_connected_message(
            {
                "type": "show_lovelace_view",
                "hassUrl": self.hass_url,
                "hassUUID": self.hass_uuid,
                "viewPath": view_path,
                "urlPath": url_path,
            },
            callback_function=callback_function,
        )

    def _send_connected_message(
        self, data: dict[str, Any], callback_function: CallbackType | None
    ) -> None:
        """Send a message to a connected Home Assistant Cast"""
        _LOGGER.debug("HomeAssistantController._send_connected_message %s", data)
        if self.hass_connected:
            _LOGGER.debug(
                "HomeAssistantController._send_connected_message already connected"
            )
            self.send_message_nocheck(data, callback_function=callback_function)
            return

        _LOGGER.debug(
            "HomeAssistantController._send_connected_message not yet connected"
        )
        self._connect_hass(
            chain_on_success(
                partial(self.send_message_nocheck, data), callback_function
            )
        )



"""
Controller to monitor audio group members.
"""

from __future__ import annotations

import abc
import logging
from typing import TYPE_CHECKING, TypedDict
from uuid import UUID

from . import BaseController
from .media import MediaController, MediaStatusListener, MediaStatus
from .receiver import CastStatus, CastStatusListener
from ..const import MESSAGE_TYPE

# pylint: disable-next=no-name-in-module
from ..generated.cast_channel_pb2 import CastMessage
from ..socket_client import (
    CONNECTION_STATUS_CONNECTED,
    CONNECTION_STATUS_DISCONNECTED,
    CONNECTION_STATUS_LOST,
    ConnectionStatus,
    ConnectionStatusListener,
)

if TYPE_CHECKING:
    from .. import Chromecast

_LOGGER = logging.getLogger(__name__)

MULTIZONE_NAMESPACE = "urn:x-cast:com.google.cast.multizone"
TYPE_CASTING_GROUPS = "CASTING_GROUPS"
TYPE_DEVICE_ADDED = "DEVICE_ADDED"
TYPE_DEVICE_UPDATED = "DEVICE_UPDATED"
TYPE_DEVICE_REMOVED = "DEVICE_REMOVED"
TYPE_GET_CASTING_GROUPS = "GET_CASTING_GROUPS"
TYPE_GET_STATUS = "GET_STATUS"
TYPE_MULTIZONE_STATUS = "MULTIZONE_STATUS"
TYPE_SESSION_UPDATED = "PLAYBACK_SESSION_UPDATED"


class GroupInfo(TypedDict):
    """Chromecast connection and listener for a group."""

    chromecast: Chromecast
    listener: Listener


class GroupMemberInfo(TypedDict):
    """Group memberships and listener for a group."""

    group_memberships: set[str]
    listeners: list[MultiZoneManagerListener]


class MultiZoneControllerListener(abc.ABC):
    """Listener for receiving audio group events."""

    @abc.abstractmethod
    def multizone_member_added(self, group_uuid: str) -> None:
        """The cast has been added to group identified by group_uuid."""

    @abc.abstractmethod
    def multizone_member_removed(self, group_uuid: str) -> None:
        """The cast has been removed from group identified by group_uuid."""

    @abc.abstractmethod
    def multizone_status_received(self) -> None:
        """Multizone status has been updated."""


class MultiZoneManagerListener(abc.ABC):
    """Listener for receiving audio group events for a cast device."""

    @abc.abstractmethod
    def added_to_multizone(self, group_uuid: str) -> None:
        """The cast has been added to group identified by group_uuid."""

    @abc.abstractmethod
    def removed_from_multizone(self, group_uuid: str) -> None:
        """The cast has been removed from group identified by group_uuid."""

    @abc.abstractmethod
    def multizone_new_media_status(
        self, group_uuid: str, media_status: MediaStatus
    ) -> None:
        """The group identified by group_uuid, of which the cast is a member, has new media status."""

    @abc.abstractmethod
    def multizone_new_cast_status(
        self, group_uuid: str, cast_status: CastStatus
    ) -> None:
        """The group identified by group_uuid, of which the cast is a member, has new status."""


class Listener(
    CastStatusListener,
    ConnectionStatusListener,
    MediaStatusListener,
    MultiZoneControllerListener,
):
    """Callback handler."""

    def __init__(
        self, group_cast: Chromecast, casts: dict[str, GroupMemberInfo]
    ) -> None:
        """Initialize the listener."""
        self._casts = casts
        group_cast.register_status_listener(self)
        group_cast.media_controller.register_status_listener(self)
        group_cast.register_connection_listener(self)
        self._mz = MultizoneController(group_cast.uuid)
        self._mz.register_listener(self)
        self._group_uuid = str(group_cast.uuid)
        group_cast.register_handler(self._mz)

    def new_cast_status(self, status: CastStatus) -> None:
        """Handle reception of a new CastStatus."""
        casts = self._casts
        group_members = self._mz.members
        for member_uuid in group_members:
            if member_uuid not in casts:
                continue
            for listener in list(casts[member_uuid]["listeners"]):
                listener.multizone_new_cast_status(self._group_uuid, status)

    def new_media_status(self, status: MediaStatus) -> None:
        """Handle reception of a new MediaStatus."""
        casts = self._casts
        group_members = self._mz.members
        for member_uuid in group_members:
            if member_uuid not in casts:
                continue
            for listener in list(casts[member_uuid]["listeners"]):
                listener.multizone_new_media_status(self._group_uuid, status)

    def load_media_failed(self, queue_item_id: int, error_code: int) -> None:
        """Called when load media failed."""

    def new_connection_status(self, status: ConnectionStatus) -> None:
        """Handle reception of a new ConnectionStatus."""
        if status.status == CONNECTION_STATUS_CONNECTED:
            self._mz.update_members()
        if status.status in (
            CONNECTION_STATUS_DISCONNECTED,
            CONNECTION_STATUS_LOST,
        ):
            self._mz.reset_members()

    def multizone_member_added(self, group_uuid: str) -> None:
        """Handle added audio group member."""
        casts = self._casts
        if group_uuid not in casts:
            casts[group_uuid] = {"listeners": [], "group_memberships": set()}
        casts[group_uuid]["group_memberships"].add(self._group_uuid)
        for listener in list(casts[group_uuid]["listeners"]):
            listener.added_to_multizone(self._group_uuid)

    def multizone_member_removed(self, group_uuid: str) -> None:
        """Handle removed audio group member."""
        casts = self._casts
        if group_uuid not in casts:
            casts[group_uuid] = {"listeners": [], "group_memberships": set()}
        casts[group_uuid]["group_memberships"].discard(self._group_uuid)
        for listener in list(casts[group_uuid]["listeners"]):
            listener.removed_from_multizone(self._group_uuid)

    def multizone_status_received(self) -> None:
        """Handle reception of audio group status."""


class MultizoneManager:
    """Manage audio groups."""

    def __init__(self) -> None:
        # Protect self._casts because it will be accessed from callbacks from
        # the casts' socket_client thread
        self._casts: dict[str, GroupMemberInfo] = {}
        self._groups: dict[str, GroupInfo] = {}

    def add_multizone(self, group_cast: Chromecast) -> None:
        """Start managing a group"""
        self._groups[str(group_cast.uuid)] = {
            "chromecast": group_cast,
            "listener": Listener(group_cast, self._casts),
        }

    def remove_multizone(self, group_uuid: UUID) -> None:
        """Stop managing a group"""
        group_uuid_str = str(group_uuid)
        group = self._groups.pop(group_uuid_str, None)
        # Inform all group members that they are no longer members
        if group is not None:
            group["listener"]._mz.reset_members()  # pylint: disable=protected-access
        for member in self._casts.values():
            member["group_memberships"].discard(group_uuid_str)

    def register_listener(
        self, member_uuid: UUID, listener: MultiZoneManagerListener
    ) -> None:
        """Register a listener for audio group changes of cast uuid.
        On update will call:
        listener.added_to_multizone(group_uuid)
            The cast has been added to group uuid
        listener.removed_from_multizone(group_uuid)
            The cast has been removed from group uuid
        listener.multizone_new_media_status(group_uuid, media_status)
            The group uuid, of which the cast is a member, has new status
        listener.multizone_new_cast_status(group_uuid, cast_status)
            The group uuid, of which the cast is a member, has new status
        """
        member_uuid_str = str(member_uuid)
        if member_uuid_str not in self._casts:
            self._casts[member_uuid_str] = {"listeners": [], "group_memberships": set()}
        self._casts[member_uuid_str]["listeners"].append(listener)

    def deregister_listener(
        self, member_uuid: UUID, listener: MultiZoneManagerListener
    ) -> None:
        """Deregister listener for audio group changes of cast uuid."""
        self._casts[str(member_uuid)]["listeners"].remove(listener)

    def get_multizone_memberships(self, member_uuid: UUID) -> list[str]:
        """Return a list of audio groups in which cast member_uuid is a member"""
        return list(self._casts[str(member_uuid)]["group_memberships"])

    def get_multizone_mediacontroller(self, group_uuid: UUID) -> MediaController:
        """Get mediacontroller of a group"""
        return self._groups[str(group_uuid)]["chromecast"].media_controller


class MultizoneController(BaseController):
    """Controller to monitor audio group members."""

    def __init__(self, uuid: UUID) -> None:
        self._members: dict[str, str] = {}
        self._status_listeners: list[MultiZoneControllerListener] = []
        self._uuid = str(uuid)
        super().__init__(MULTIZONE_NAMESPACE, target_platform=True)

    def _add_member(self, uuid: str, name: str) -> None:
        if uuid not in self._members:
            self._members[uuid] = name
            _LOGGER.debug(
                "(%s) Added member %s(%s), members: %s",
                self._uuid,
                uuid,
                name,
                self._members,
            )
            for listener in list(self._status_listeners):
                listener.multizone_member_added(uuid)

    def _remove_member(self, uuid: str) -> None:
        name = self._members.pop(uuid, "<Unknown>")
        _LOGGER.debug(
            "(%s) Removed member %s(%s), members: %s",
            self._uuid,
            uuid,
            name,
            self._members,
        )
        for listener in list(self._status_listeners):
            listener.multizone_member_removed(uuid)

    def register_listener(self, listener: MultiZoneControllerListener) -> None:
        """Register a listener for audio group changes. On update will call:
        listener.multizone_member_added(uuid)
        listener.multizone_member_removed(uuid)
        listener.multizone_status_received()
        """
        self._status_listeners.append(listener)

    @property
    def members(self) -> list[str]:
        """Return a list of audio group members."""
        return list(self._members.keys())

    def reset_members(self) -> None:
        """Reset audio group members."""
        for uuid in list(self._members):
            self._remove_member(uuid)

    def update_members(self) -> None:
        """Update audio group members."""
        self.send_message({MESSAGE_TYPE: TYPE_GET_STATUS})

    def get_casting_groups(self) -> None:
        """Send GET_CASTING_GROUPS message."""
        self.send_message({MESSAGE_TYPE: TYPE_GET_CASTING_GROUPS})

    def receive_message(  # pylint: disable=too-many-return-statements
        self, _message: CastMessage, data: dict
    ) -> bool:
        """Called when a multizone message is received."""
        if data[MESSAGE_TYPE] == TYPE_DEVICE_ADDED:
            uuid = data["device"]["deviceId"]
            name = data["device"]["name"]
            self._add_member(uuid, name)
            return True

        if data[MESSAGE_TYPE] == TYPE_DEVICE_REMOVED:
            uuid = data["deviceId"]
            self._remove_member(uuid)
            return True

        if data[MESSAGE_TYPE] == TYPE_DEVICE_UPDATED:
            uuid = data["device"]["deviceId"]
            name = data["device"]["name"]
            self._add_member(uuid, name)
            return True

        if data[MESSAGE_TYPE] == TYPE_MULTIZONE_STATUS:
            members = data["status"]["devices"]
            members = {member["deviceId"]: member["name"] for member in members}
            removed_members = list(set(self._members.keys()) - set(members.keys()))
            added_members = list(set(members.keys()) - set(self._members.keys()))
            _LOGGER.debug(
                "(%s) Added members %s, Removed members: %s",
                self._uuid,
                added_members,
                removed_members,
            )

            for uuid in removed_members:
                self._remove_member(uuid)
            for uuid in added_members:
                self._add_member(uuid, members[uuid])

            for listener in list(self._status_listeners):
                listener.multizone_status_received()

            return True

        if data[MESSAGE_TYPE] == TYPE_SESSION_UPDATED:
            # A temporary group has been formed
            return True

        if data[MESSAGE_TYPE] == TYPE_CASTING_GROUPS:
            # Answer to GET_CASTING_GROUPS
            return True

        return False

    def tear_down(self) -> None:
        """Called when controller is destroyed."""
        super().tear_down()

        self._status_listeners = []
